package org.sanskritlibrary.viewer;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.sanskritlibrary.Annotation;
import org.sanskritlibrary.LineInfo;
import org.sanskritlibrary.ManuscriptInfo;
import org.sanskritlibrary.TextFormatter;
import org.sanskritlibrary.TextStructure;
import org.sanskritlibrary.servlet.SanskritLibraryServlet;

/**
 *
 * @author ralph
 */
public class LoadImage extends SanskritLibraryServlet {
   
    //static String mss = "/Users/ralph/SanskritLibrary/LatestGrantFiles/mss";
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        String imageFile = null;
        String mss = (String)this.getServletContext().getInitParameter("dataDirectory") + "mss";

        String type = request.getParameter("type");
        if (type != null) {
            String manuscript = null;
            String dataDirectory = null;
            String textDirectory = null;
            String tsfn = null;
            int lineNumber = 0;
            int dirnum = 0;
            TextStructure ts = null;
            LineInfo li = null;
            String text = request.getParameter("text");
            if (text != null) {
                manuscript = request.getParameter("manuscript");
                lineNumber = Integer.parseInt(request.getParameter("line"));
                dataDirectory = (String)this.getServletContext().getInitParameter("dataDirectory");
                textDirectory =  dataDirectory + "texts";
                tsfn = String.format("%s/%s.bin", textDirectory, text);
                try {
                    ts = getTextStructure(request, tsfn);
                } catch (Exception ex) {
                    throw new ServletException(ex.getMessage());
                }
                for (ManuscriptInfo mi : ts.manuscriptDirectories) {
                    if (mi.directoryName.equals(manuscript)) {
                        break;
                     } else {
                        dirnum++;
                     }
                }
                li = ts.lineInfo.get(lineNumber);
            }

            if (type.equals("cropped")) {
                Config config = (Config)this.getServletContext().getAttribute("config");
                String file = request.getParameter("file");         // catalog file, e.g. penn0492
                String img = request.getParameter("image");         // image file, e.g., mscoll390_item492_wk1_body0001_crop1
                String fn = config.filemap.get(file);
                if (fn == null) {
                    PrintWriter pw = response.getWriter();
                    pw.write(String.format("%s not found in config.xml", file));
                    return;
                } else {
                    imageFile = String.format("%s/%s/%s.jpg", mss, fn, img);
                }
            } else if (type.equals("sita")) {

                try {
                    Annotation a = li.annotations.get(dirnum);
                    imageFile = TextFormatter.completeName(mss, manuscript, a.image, ts.manuscriptDirectories.get(dirnum).prefixes);
                } catch (Exception ex) {
                    throw new IOException(ex.getMessage());
                }
            } else if (type.equals("folio")) {
                Annotation a = li.annotations.get(dirnum);
                if (a.image == null) {
                    return;
                }
                String image = a.image;
                int len = ts.lineInfo.size();
                int i = 0;
                // Find first annotation on image
                for (; i<len; i++) {
                    a = ts.lineInfo.get(i).annotations.get(dirnum);
                    if (a != null && a.image.equals(image)) {
                        break;
                   }
                }

                StringBuilder list = new StringBuilder("[");
                String delim = "";
                // Add annotation information until last annotation for image reached.
                for (; i<len; i++) {
                    a = ts.lineInfo.get(i).annotations.get(dirnum);
                    if (a != null) {
                        if (!a.image.equals(image))
                        {
                            break;      // we have all annotations for image now
                        } else {
                            list.append(String.format("%s{\"id\":%d, \"x1\":%d, \"y1\":%d, \"x2\":%d, \"y2\":%d}",
                                    delim, i, a.x1, a.y1, a.x2, a.y2));
                            delim = "\n,";
                        }
                    }
                }
                list.append("]");
                PrintWriter pw = response.getWriter();
                pw.write(list.toString());
                return;
            }
        } else {
            String name = request.getParameter("name");
            if (name == null) {
                PrintWriter pw = response.getWriter();
                pw.write(String.format("Parameter name is null"));
                return;
            } else {
                imageFile = String.format("%s/%s", mss, name);
            }
        }

        try {
            FileInputStream fis = new FileInputStream(imageFile);
            byte[] b = new byte[fis.available()];
            fis.read(b);
            ServletOutputStream sos = response.getOutputStream();
            sos.write(b);
            response.setContentType("image/jpeg");
        } catch (Exception ex) {
            sendMessage(response, ex.getMessage());
        }
    }




    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
    }


    private void sendMessage(HttpServletResponse response, String msg) throws IOException {
        response.setContentType("text/html; charset=UTF-8");
        PrintWriter pw = response.getWriter();
        pw.write(msg);
    }

    /** 
     * Returns a short description of the servlet.
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "Short description";
    }// </editor-fold>

}
