/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.sanskritlibrary.viewer;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.servlet.SanskritLibraryServlet;
import org.sanskritlibrary.Regex;

/**
 *
 * @author ralph
 */
public class OCRHelper extends SanskritLibraryServlet {


    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        String ocrPrep = getServletContext().getInitParameter("OCRPrep");
        if (!ocrPrep.endsWith("/")) {
            ocrPrep += "/";
        }        
        String mappings = String.format("%smappings/", ocrPrep);
        File f = new File(mappings);
        f.mkdirs();
        String xml = this.getServletContext().getRealPath("ocr.xml");
        String html = this.getServletContext().getRealPath("ocr.html");
        StringBuilder sb = new StringBuilder();

        String command = request.getParameter("command");
        if (command == null) {
            command = "get";
        }
        
        if (command.equals("loadImage")) {
            try {
                String image = request.getParameter("image");
                InputStream is = new FileInputStream(String.format("%s/%s.jpg", ocrPrep, image));
                int len = is.available();
                byte[] b = new byte[len];
                is.read(b);
                OutputStream os = response.getOutputStream();
                os.write(b);
            } catch (Exception ex) {
                throw new IOException(ex.getMessage());
            }
        } else if (command.equals("getState")) {            
          try {
              String manuscriptName = request.getParameter("manuscriptName");
              InputStream is = new FileInputStream(String.format("%s%s.txt", mappings, manuscriptName));
              int len = is.available();
              byte[] b = new byte[len];
              is.read(b);              
              OutputStream os = response.getOutputStream();
              response.setContentType("text/text");
              os.write(b);              
          } catch (Exception ex) {
              throw new IOException(ex.getMessage());
          }
        } else if (command.equals("saveState")) {
              String manuscriptName = request.getParameter("manuscriptName");
              String state = request.getParameter("state");
              OutputStream is = new FileOutputStream(String.format("%s%s.txt", mappings, manuscriptName));
              is.write(state.getBytes("UTF-8"));
              is.close();
              
              OutputStream os = response.getOutputStream();              
              response.setContentType("text/text");
              os.write("OK".getBytes("UTF-8"));                          
        } else {
            try {
                File test = new File(ocrPrep + "rebuildMappings.txt");      // if this file exists, regenerate mappings
                boolean rebuildMappings = false;
                if (!test.exists()) {
                    rebuildMappings = true;
                    test.createNewFile();
                }
                
                if (rebuildMappings) {
                    Hashtable<String, Integer> h = new Hashtable<String, Integer>();
                    getLineCounts(ocrPrep, h);
                    ArrayList<String> a = new ArrayList<String>();
                    Enumeration<String> en = h.keys();
                    while (en.hasMoreElements()) {
                        a.add(en.nextElement());
                    }
                    Collections.sort(a);
                    for (String s : a) {
                        System.out.format("%s: %d\n", s, h.get(s));
                    }
                    String templateHTML = this.getServletContext().getRealPath("template.html");
                    ParseOCRPrep pop = new ParseOCRPrep(xml, 
                            html, 
                            this.getServletContext().getRealPath("manuscripts"), 
                            h, 
                            templateHTML);
                    pop.setOCRPrepDirectory(ocrPrep);
                    pop.parse();
                }

                String s = IOAssistance.readEntireFile(html);
                this.writeResult(response, s, "text/html");
            } catch (Exception ex) {
                throw new ServletException(ex);
            }
        }
    }

    
    private void getLineCounts(String dirname, Hashtable<String, Integer> h) throws IOException {
        Regex re = new Regex("", "^(.*)_line(\\d+)\\.jpg$");
        File dir = new File(dirname);
        if (!dir.exists()) {
            throw new IOException(String.format("Value of OCRPrep in Tomcat's context.xml is %s. This directory does not exist.",
                    dirname));
        }
        File[] files = dir.listFiles();
        for (File f : files) {
            if (f.isDirectory()) {
                continue;
            }
            
            String fn = f.getName();
            if (fn.endsWith(".jpg")) {
                re.reset(fn);
                if (re.find()) {
                    String key = re.getGroup(1);
                    int count = 0;
                    if (h.get(key) != null) {
                        count = h.get(key).intValue();
                    }
                    h.put(key, new Integer(++count));
                } else {
                    throw new IOException(String.format("Unexpected file name %s in %s. Should match pattern ^(.*)_line(\\d+)\\.jpg$"));
                }
            }
        }
    }

    
    private long getTimestamp(String fn) {
	File file = new File(fn); 
        if (file.exists()) {
            return file.lastModified();
        } else {
            return 0L;          // make it look like the file is very, very old.
        }
    }

    
    private void display(StringBuilder sb, String fn) {
	File file = new File(fn); 
	sb.append(String.format("Before Format : %s<br/>", Long.toString(file.lastModified())));
	SimpleDateFormat sdf = new SimpleDateFormat("MM/dd/yyyy HH:mm:ss"); 
	sb.append(String.format("After Format : %s<br/>", sdf.format(file.lastModified())));                
    }
    
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
    }

    /**
     * Returns a short description of the servlet.
     *
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "Short description";
    }// </editor-fold>
}
