package org.sanskritlibrary.viewer;

// This does nothing. It is a template for writing a class that parses XML

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
//DefaultHandler implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler
//import org.xml.sax.helpers.DefaultHandler;

//DefaultHandler2 implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler, DeclHandler, EntityResolver2, LexicalHandler
//http://docs.oracle.com/javase/1.5.0/docs/api/org/xml/sax/ext/DefaultHandler2.html
import org.xml.sax.ext.DefaultHandler2;         // this allows processing of comments because it implements LexicalHandler


// SAXParserHandler is an interface that forces us to include required methods.
// Serializable is optional. Have made local variables that don't need to be serialized transient.
public class ParseOCRPrep extends DefaultHandler2 implements SAXParserHandler, Serializable {
    transient Locator locator;
    transient ArrayList<String> pathStack = new ArrayList<String>();
    transient Hashtable<String, StringBuilder> buffers = new Hashtable<String, StringBuilder>();
    transient Regex re = new Regex("", "/(.*)\\.jpg$");
    private StringBuilder sbBody = new StringBuilder();
    private StringBuilder sbIndex = new StringBuilder();
    private StringBuilder sbErrors = new StringBuilder();
    private StringBuilder sbState = new StringBuilder();
    private StringBuilder sbLineNumbers = new StringBuilder();
    private String delimLineNumbers = "";
    private String delimIndex = "";
    private StringBuilder sbTranscriptions = new StringBuilder();
    private String delimScript = "";
    private String xmlFile;
    private String htmlFile;
    private String manuscriptsDirectory;
    private String templateHTML;
    private Hashtable<String, Integer> h;
    private Hashtable<String, Integer> hDuplicate = new Hashtable<String, Integer>();
    private BufferedWriter bwMainPage;
    private BufferedWriter bwManuscriptPage;
    private String currentJPG;
    private String currentImageKey;
    private int currentImageCount = 0;
    private int currentOverflowIndex = 0;
    private int currentLineCount = 0;
    private int currentMaxLine = 0;
    private int unknownSuffix = 1;
    private ArrayList<LineInfo> lines = new ArrayList<LineInfo>();
    private LineInfo currentLineInfo;
    private String currentCatalogName;
    private String ocrPrepDirectory;
    private int currentCropNumber = 0;
    StringBuilder badLines = new StringBuilder();
    boolean incompleteLinesFound = false;
    boolean duplicateLineFound = false;
    StringBuilder sbMissingImages = new StringBuilder();
    
    public ParseOCRPrep(String xmlFile, String htmlFile, String manuscriptsDirectory, Hashtable<String, Integer> h, String templateHTML) throws Exception
    {
        this.xmlFile = xmlFile;
        this.htmlFile = htmlFile;
        this.manuscriptsDirectory = manuscriptsDirectory;
        this.h = h;
        this.templateHTML = IOAssistance.readEntireFile(templateHTML);
    }
    
    public void setOCRPrepDirectory(String ocrPrepDirectory) {
        this.ocrPrepDirectory = ocrPrepDirectory;
    }
    
    public void parse () throws Exception {
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            bwMainPage = IOAssistance.openWriter(htmlFile);
            File file = new File(xmlFile);
            InputStream inputStream= new FileInputStream(file);
            int b;
            int count = 0;
            while (true) {
                b = inputStream.read();
                if (b == '<') {
                    break;
                }
                count++;
            }
            inputStream.close();
            inputStream= new FileInputStream(file);
            for (int i=0; i<count; i++) {
                inputStream.read();
            }
 
            saxParser.parse(inputStream, this);            
            //saxParser.parse( new File(xmlFile), this);
        } catch (Exception ex) {
            throw ex;
        }
    }    

	// Maintains a stack of element names (qName)
    private void push (String qName) {
        pathStack.add(qName);
    }

	// Pop a qName from the stack
    private String pop () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
            pathStack.remove(last);
        }

        return top;
    }

	// Peek at a qName without popping stack
    private String peek () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
        }

        return top;

    }

	// Serialize the qName stack. The string that is returned represent the path to the current qName
    private String buildPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : pathStack) {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }
        return sb.toString();
    }

    private void displayError(String msg) {
        System.out.format("Line %d: %s\n", locator.getLineNumber(), msg);
    }

    @Override
    public void warning(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void error(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    @Override
    public void startDocument() throws SAXException {
        try {
            bwMainPage.write("<html>\n<head>\n<title>OCR Prep</title>\n</head>\n<body>\n");
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    @Override
    public void endDocument() throws SAXException {
        try {
            bwMainPage.write("</body>\n");
            bwMainPage.close();
        } catch (Exception ex) {
            throw new SAXException(ex);            
        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName);
        if (locator.getLineNumber() == 320) {
            localName = null;
        }
        try {
            if (qName.equals("manuscript")) {
                unknownSuffix = 1;
                String catalogName = atts.getValue("catalog");
                String shtml = catalogName.replace(".xml", ".html");
                currentCatalogName = catalogName.replace(".xml", "");                
                bwMainPage.write(String.format("<a target=\"_blank\" href=\"manuscripts/%s\">%s</a><br/>\n",shtml, currentCatalogName));
                bwManuscriptPage = IOAssistance.openWriter(String.format("%s/%s", manuscriptsDirectory, shtml));
                sbBody.setLength(0);
                
                sbErrors.setLength(0);
                sbBody.append("<div id=\"errors\">");       // ending </div> will come later
                
                sbIndex.setLength(0);
                sbIndex.append("\n");      // so that %s in template can be //%s
                sbIndex.append("var index=[\n");
                
                sbTranscriptions.setLength(0);
                sbTranscriptions.append("\n");      // so that %s in template can be //%s
                
                sbState.setLength(0);
                sbLineNumbers.setLength(0);
                
                delimIndex = "";
                delimScript = "";
                delimLineNumbers = "";
                currentCropNumber = 0;
            } else if (qName.equals("image")) {
                incompleteLinesFound = false;
                duplicateLineFound = false;
                currentJPG = atts.getValue("name");
                if (currentJPG.equals("mscoll390_item351/mscoll390_item351_wk1_body0015_crop1.jpg")) {
                    localName = null;
                }
                Integer cnt = hDuplicate.get(currentJPG);
                if (cnt != null) {
                    hDuplicate.put(currentJPG, new Integer(cnt.intValue() + 1));
                } else {
                    hDuplicate.put(currentJPG, new Integer(1));
                }
                re.reset(currentJPG);
                if (re.find()) {
                    currentImageKey = re.getGroup(1);
                } else {
                    currentImageKey = String.format("%s_%d", currentJPG, unknownSuffix++);
                }
                currentLineCount = 0;
                currentImageCount = 0;
                currentMaxLine = 0;
                Integer s = h.get(currentImageKey);
                if (s != null) {
                    currentImageCount = s.intValue();
                }
                
                currentOverflowIndex = currentImageCount + 1;   // since line numbers are 1-based in ocr.xml, they are converted to 0-based when sbLineNumbers is created.

                if (currentImageCount > 0) {
                    currentCropNumber++;
                    sbIndex.append(String.format("%s{transcriptions:%s, numSplits:%d, nextOverflowID:%d}", 
                            delimIndex, currentImageKey, currentImageCount, currentImageCount+1));
                    sbTranscriptions.append(String.format("var %s=[\n", currentImageKey));
                    if (sbState.length() > 0) {
                        sbState.append(sbLineNumbers.toString());
                        sbLineNumbers.setLength(0);
                        sbState.append("\n");
                    }
                    sbState.append(String.format("%s:", currentImageKey));
                    delimIndex = ",\n";
                    delimScript = "";
                    delimLineNumbers = "";
                } else {
                    sbErrors.append(String.format("<p style=\"color:red\">No splits for %s, transcriptions ignored.</p>\n",
                            currentImageKey));
                    this.sbMissingImages.append(String.format("%s\n", currentImageKey));
                }
            } else if (qName.equals("line")) {
                currentLineCount++;
                currentLineInfo = new LineInfo(atts.getValue("n"));
                if (currentLineInfo.n > currentMaxLine) {
                    currentMaxLine = currentLineInfo.n;
                }                
                lines.add(currentLineInfo);
                currentLineInfo.corrected = atts.getValue("corrected") != null;
                currentLineInfo.incomplete = atts.getValue("incomplete") != null;
                currentLineInfo.duplicate = atts.getValue("duplicate") != null;
            }
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
            String path = buildPath();
            StringBuilder buffer = buffers.get(path);
            String value = "";
            if (buffer != null) {
                value = buffer.toString().trim();
                buffer.setLength(0);
            }

            if (qName.equals("line")) {
                // note: if currentImageCount == 0 this information will not be written to the HTML file.
                if (currentImageCount > 0) {
                    currentLineInfo.transcription = value;
                    sbTranscriptions.append(String.format("%s{n:%d, splitNumber:-1, complete:%s, text:\"%s\"}",
                            delimScript,
                            currentLineInfo.n,
                            currentLineInfo.incomplete ? "false" : "true",
                            currentLineInfo.transcription));
                    delimScript = ",\n";
                    int ln = currentLineInfo.n;
                    if (currentLineInfo.incomplete) {
                        ln = currentOverflowIndex++;
                    } else if (currentLineInfo.n > currentImageCount) {
                        ln = currentOverflowIndex++;
                    } else if (currentLineInfo.duplicate) {
                        ln = currentOverflowIndex++;
                    }
                    sbLineNumbers.append(String.format("%s%d", delimLineNumbers, ln - 1));
                    delimLineNumbers = ",";
                }
            } else if (qName.equals("image")) {
                if (currentImageCount > 0) {
                    sbBody.append(String.format("<h2>%s</h2>\n", currentJPG));
                    sbBody.append(String.format("<button id=\"remove.%d\" onclick=\"removeAll(this)\" style=\"font-size:24px\">Do not use for OCR training</button><br/><br/>\n", currentCropNumber));
                    for (int i = 0; i < lines.size(); i++) {
                        if (!incompleteLinesFound && lines.get(i).incomplete) {
                            sbBody.append("<p style='color:red'>Note: incomplete lines are in transcription pool</p>\n");
                            incompleteLinesFound = true;
                        }
                        if (!duplicateLineFound && lines.get(i).duplicate) {
                            sbBody.append("<p style='color:red'>Note: duplicate line(s) are in transcription pool</p>\n");
                            duplicateLineFound = true;
                        }
                    }
                    if (currentLineCount > currentImageCount) {
                        if (currentImageCount > 0) {
                            badLines.append(String.format("new ImageInfo(\"%s\", %d),  // missing lines?\n",
                                    currentJPG, currentLineCount));
                            sbBody.append(String.format("<p style='color:red'>Note: overflow lines are in transcription pool</p>", currentJPG));
                        }
                    } else if (currentMaxLine > currentImageCount) {
                            badLines.append(String.format("new ImageInfo(\"%s\", %d),  // side lines?\n",
                                    currentJPG, currentMaxLine));
                        sbBody.append(String.format("<p style='color:red'>Note: overflow lines are in transcription pool</p>", currentJPG));
                    }
                    sbTranscriptions.append("\n];\n");
                    if (h.get(currentImageKey) != null) {
                        currentImageCount = h.get(currentImageKey);

                        sbBody.append("<table border=\"1px\" style=\"border-collapse:collapse\">\n");
                        for (int i = 0; i < currentImageCount; i++) {      // imageCount is number of lines in the pre-split manuscript image.
                            sbBody.append("<tr>\n");
                            String upButton, downButton, outButton;
                            if (i != 0) {
                                upButton = String.format("<button id=\"%s\" onclick=\"moveLinesUp(this)\">up</button>",
                                        buildID("up", i));
                            } else {
                                upButton = "";
                            }

                            if (i != currentImageCount - 1) {
                                downButton = String.format("<button id=\"%s\" onclick=\"moveLinesDown(this)\">down</button>",
                                        buildID("down", i));
                            } else {
                                downButton = "";
                            }

                            outButton = String.format("<button id=\"%s\" onclick=\"moveOut(this)\">out</button>",
                                    buildID("out", i));
                            sbBody.append(String.format("<td>%d.</td>\n", i + 1));
                            sbBody.append(String.format("<td><div id=\"%s\">%s<br/>%s<br/>%s</div></td>\n",
                                    buildID("div", i), upButton, downButton, outButton));
                            sbBody.append(String.format("<td><img id='%s' src=\"/ManuscriptViewer/OCRHelper?command=loadImage&image=%s_line%d\" height=\"20%s\"/><br><p id='x%d.%d'></p></td>\n",
                                    String.format("image%d.%d", currentCropNumber, i),
                                    currentImageKey, i + 1, "%", currentCropNumber, i + 1));
                            sbBody.append("</tr>\n");
                        }
                        sbBody.append("</table>\n");
                        sbBody.append(String.format("<div id='%s' style='border:2px solid red'></div>\n",
                                "pool." + this.currentCropNumber));
                    }
                }
                // else don't create array of transcriptions for current image because Donglai didn't us any splits.

                lines.clear();
            } else if (qName.equals("condition")) {
                sbBody.append(String.format("<p>%s</p>\n", value));
            } else if (qName.equals("manuscript")) {
                sbIndex.append("];\n");
                sbErrors.append("</div>");
                bwManuscriptPage.write(String.format(templateHTML, 
                        currentCatalogName,         // placeholder: title
                        currentCatalogName,         // placeholder: manuscript name
                        sbTranscriptions.toString(),        // placeholder: transcriptions
                        sbIndex.toString(),         // placeholder: image index (points to above transcriptions)
                        sbErrors.toString(),        // placeholder: errors div
                        sbBody.toString()));        // placeholder: images and transcriptions

                bwManuscriptPage.close();
                
                BufferedWriter bwState = IOAssistance.openWriter(String.format("%smappings/%s.txt", 
                        ocrPrepDirectory,
                        currentCatalogName));
                bwState.write(sbState.toString());
                bwState.write(sbLineNumbers.toString());   // since this is done on a startElement for "image"
                sbState.setLength(0);
                sbLineNumbers.setLength(0);
                bwState.close();
            }
        } catch (Exception ex) {
            throw new SAXException(ex.getMessage());
        } finally {
            pop();
        }
    }

    private String buildID(String type, int i) {
        return String.format("image.%s.%s%s", currentCropNumber, type, i+1);
    }

    @Override
    // Collects the innerText in a StringBuilder keyed by the current path. It is endElement's responsibility
    // to retrieve the StringBuilder and set its length to 0.
    // This is optional. It is not specified in SAXParserHandler
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder sb = buffers.get(path);
        if (sb == null) {
            sb = new StringBuilder();
            buffers.put(path, sb);
        }
        sb.append(new String(ch, start, length));
    }

    @Override
    // Optional if you want to be notified of comments
    public void comment(char[] ch, int start, int length) throws SAXException {

    }
    
    class LineInfo {
        int n;
        boolean corrected = false;
        boolean incomplete = false;
        boolean duplicate = false;
        String transcription;
        LineInfo(String n) {
            this.n = Integer.parseInt(n);
        }
    }

}
